unit ScriptEditor;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, SynEdit, SynCompletion, mbOfficeColorDialog,
  Forms, Controls, Graphics, Dialogs, Menus, ActnList,
  ComCtrls, ExtCtrls, IniPropStorage, Types, LCLType, constants;

type

  TPixel = record
    Bright : integer;
    Fade : integer;
    Color : TColor;
  end;

  TStackRec = record
    NumStr : integer;
    PosStr : integer;
  end;

  { TScript }

  TScript = class(TForm)
    ActExit: TAction;
    ActDbgStartStop: TAction;
    ActDbgStop: TAction;
    ActDbgStep: TAction;
    ActAutoRun: TAction;
    ActColor: TAction;
    ActRunTo: TAction;
    ActOpen: TAction;
    ActSaveAs: TAction;
    ActSave: TAction;
    ActNew: TAction;
    ActionList: TActionList;
    AppProp: TApplicationProperties;
    Images: TImageList;
    ColorDlg: TmbOfficeColorDialog;
    PropStorage: TIniPropStorage;
    MainMenu: TMainMenu;
    MenuFile: TMenuItem;
    MenuDebug: TMenuItem;
    MenuItem1: TMenuItem;
    MenuItem10: TMenuItem;
    MenuItem11: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem5: TMenuItem;
    MenuItem6: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    MenuItem9: TMenuItem;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    ScriptEditor: TSynEdit;
    ScriptCompletion: TSynCompletion;
    StatusBar: TStatusBar;
    RunTimer: TTimer;
    ToolBar: TToolBar;
    ToolButton1: TToolButton;
    ToolButton10: TToolButton;
    ToolButton11: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ToolButton7: TToolButton;
    RunSpeed: TTrackBar;
    ToolButton8: TToolButton;
    ToolButton9: TToolButton;
    procedure ActAutoRunExecute(Sender: TObject);
    procedure ActColorExecute(Sender: TObject);
    procedure ActDbgStartStopExecute(Sender: TObject);
    procedure ActDbgStepExecute(Sender: TObject);
    procedure ActDbgStopExecute(Sender: TObject);
    procedure ActExitExecute(Sender: TObject);
    procedure ActionListUpdate(AAction: TBasicAction; var Handled: Boolean);
    procedure ActNewExecute(Sender: TObject);
    procedure ActOpenExecute(Sender: TObject);
    procedure ActRunToExecute(Sender: TObject);
    procedure ActSaveAsExecute(Sender: TObject);
    procedure ActSaveExecute(Sender: TObject);
    procedure AppPropHint(Sender: TObject);
    procedure AppPropShowHint(var HintStr: string; var CanShow: Boolean;
      var HintInfo: THintInfo);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormShowHint(Sender: TObject; HintInfo: PHintInfo);
    procedure ColorPickDblClick(Sender: TObject);
    procedure RunSpeedChange(Sender: TObject);
    procedure RunTimerTimer(Sender: TObject);
    procedure ScriptEditorSpecialLineColors(Sender: TObject; Line: integer;
      var Special: boolean; var FG, BG: TColor);
  private
    { private declarations }
    DbgMode : boolean;
    CurStr : integer;
    CurPos : integer;
    ToStr : integer;
    ScriptFilename : string;
    Stack : array[0 .. cStackDepth-1] of TStackRec;
    StackHead : integer;

    TimerWait : integer;
    DbgInd : integer;

    procedure SelectStr(SN : integer);
    function CheckAndSave : boolean;
    procedure DebugMode(const Enable : boolean);
    function DebugStep : string;
    procedure UpdateFilename;
    procedure LoadFile(const FN : string);
    procedure StatusShort(const S : string);
    procedure StatusLong(const S : string);
    procedure StatusRun(const S : string);
    procedure ExecCmd(S : string);
    function GetByte(const Cmd: string): integer;
    function GetColor(const Cmd: string): Integer;
    function GetNum(const Ch: Char): integer;
    procedure UpdateDbgViews;
    function SatAdd(X, Y : integer):integer;
    function SatSub(X, Y : integer):integer;
    function SatMul(X, Y : integer):integer;

    procedure PushStack(aStr : integer; aPos : integer = 0); overload;
    procedure PushStack(aST : TStackRec); overload;
    function PopStack : TStackRec;

    procedure Wait(WT : integer);
    procedure Fade;
    procedure DbgEnd;
    procedure SetRunSpeed;
    function ColorToText(aColor : TColor):string;
    procedure DebugAutoIndicator;
  public
    CurV : array[0 .. cTotalVar-1] of integer;
    CurColor : TColor;
    CurFade : integer;
    CurBright : integer;
    Rev : boolean;
    Pixel : array[0..cTotalPix-1] of TPixel;
    { public declarations }
  end;

var
  Script: TScript;

implementation
uses DbgView;
{$R *.frm}

procedure TScript.ActExitExecute(Sender: TObject);
begin
  if CheckAndSave then Close;
end;

procedure TScript.ActionListUpdate(AAction: TBasicAction; var Handled: Boolean);
begin
  ActDbgStartStop.Enabled := (Trim(ScriptEditor.Text) <> '');
  ActDbgStartStop.Checked := DbgMode;
  if DbgMode
    then ActDbgStartStop.Caption := 'Stop'
    else ActDbgStartStop.Caption := 'Start';
  ActAutoRun.Enabled := (not RunTimer.Enabled) and DbgMode;
  ActDbgStop.Enabled := RunTimer.Enabled;
  ActDbgStep.Enabled := DbgMode and (not RunTimer.Enabled);
  ActRunTo.Enabled := DbgMode and (not RunTimer.Enabled);

  ActSave.Enabled := ScriptEditor.Modified;
  ActSaveAs.Enabled := ActDbgStartStop.Enabled;
  if DbgMode
     then if RunTimer.Enabled
             then StatusShort('Running')
             else StatusShort('Debug')
     else StatusShort('Edit');
  RunSpeed.Visible := RunTimer.Enabled;//DbgMode;
  //ColorPick.Enabled := not DbgMode;
  ActColor.Enabled := not DbgMode;
  //ColorPick.Enabled := not DbgMode;
  //ActColor.Checked := ColorPick.Visible;
end;

function TScript.CheckAndSave : boolean;
begin
  Result := true;
  if(ScriptEditor.Modified) then begin
    case MessageDlg('Warning',
                    'Current script nod saved!'#$0d#$0a'Save before continue?',
                    mtWarning, mbYesNoCancel, 0) of
    mrYes : ActSave.Execute;
    mrNo : exit;
    mrCancel : Result := false;
    end;
  end;
end;

procedure TScript.ActNewExecute(Sender: TObject);
begin
  if not CheckAndSave then exit;
  ScriptFilename := '';
  UpdateFilename;
  ScriptEditor.ClearAll;
  ScriptEditor.Gutter.MarksPart;
end;

procedure TScript.LoadFile(const FN : string);
begin
  try
     ScriptEditor.Lines.LoadFromFile(FN);
  finally
     ScriptEditor.Modified := false;
     ScriptFilename := FN;
     UpdateFilename;
  end;
end;

procedure TScript.ActOpenExecute(Sender: TObject);
begin
  //OpenDialog.Title := 'Открыть файл';
  if OpenDialog.InitialDir = ''
     then OpenDialog.InitialDir := ExtractFilePath(ParamStr(0));
  if not CheckAndSave then exit;
  if OpenDialog.Execute
     then LoadFile(OpenDialog.FileName);
end;

procedure TScript.ActRunToExecute(Sender: TObject);
begin
  ToStr := ScriptEditor.CaretY;
  SetRunSpeed;
  RunTimer.Enabled := true;
end;

procedure TScript.ActSaveAsExecute(Sender: TObject);
begin
  //SaveDialog.Title := 'Сохранить файл';
  SaveDialog.InitialDir := ExtractFilePath(ScriptFilename);
  SaveDialog.FileName := ExtractFileName(ScriptFilename);
  if SaveDialog.Execute then begin
     ScriptFilename := SaveDialog.FileName;
     ActSave.Execute;
  end;
end;

procedure TScript.ActSaveExecute(Sender: TObject);
begin
  if ScriptFilename = '' then begin
    ActSaveAs.Execute;
    exit;
  end;
  try
     ScriptEditor.Lines.SaveToFile(ScriptFilename);
  finally
     ScriptEditor.Modified := false;
  end;
  UpdateFilename;
end;

procedure TScript.AppPropHint(Sender: TObject);
begin
   StatusLong(Application.Hint);
end;

procedure TScript.AppPropShowHint(var HintStr: string; var CanShow: Boolean;
  var HintInfo: THintInfo);
begin
  CanShow := false;
end;

procedure TScript.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  CanClose := CheckAndSave;
end;

procedure TScript.UpdateFilename;
var
  P, S : string;
begin
  S := ScriptFilename;
  if length(S) > 25 then begin
    S := ExtractFileName(ScriptFilename);
    P := ExtractFilePath(ScriptFilename) + '\';
    S := copy(P, 1, 10) + ' ... ' + copy(P, length(P)-5, 5) + S;
  end;
  if S = ''
     then Caption := 'Script [noname]'
     else Caption := 'Script [' + S + ']';
end;

procedure TScript.FormCreate(Sender: TObject);
begin
  Randomize;
  SaveDialog.Filter := OpenDialog.Filter;
  if ParamStr(1) <> ''
     then LoadFile(ParamStr(1))
     else ScriptFilename := '';
  UpdateFilename;
  StatusShort('Edit');
end;

procedure TScript.FormShowHint(Sender: TObject; HintInfo: PHintInfo);
begin
  StatusLong(Application.Hint);
end;

procedure TScript.ColorPickDblClick(Sender: TObject);
begin
  //ScriptEditor.InsertTextAtCaret(ColorToText(ColorPick.SelectedColor));
end;

procedure TScript.RunSpeedChange(Sender: TObject);
begin
  SetRunSpeed;
  case RunSpeed.Position of
  0: RunSpeed.Hint := 'Auto-execute: 1 step per 1 second';
  1: RunSpeed.Hint := 'Auto-execute: 2 steps per 1 second';
  2: RunSpeed.Hint := 'Auto-execute: 5 steps per 1 second';
  3: RunSpeed.Hint := 'Auto-execute: 10 steps per 1 second';
  4: RunSpeed.Hint := 'Auto-execute on maximum speed, CPU limited';
  end;
  StatusLong(RunSpeed.Hint);
end;

procedure TScript.Wait(WT: integer);
begin
  if RunTimer.Enabled then begin
    TimerWait := WT;
    exit;
  end;
  DbgForm.WaitDbg;
  while (WT <> 0) do begin
    dec(WT);
    Fade;
  end;
end;

function SatSub1(X, Y : integer) : integer;
begin
   Result := X - Y;
   if Result < 0 then Result := 0;
end;

procedure TScript.Fade;
var
  i : integer;
begin
  for i := 0 to cTotalPix-1 do
    if Pixel[i].Fade <= $7F
      then Pixel[i].Bright := SatSub1(Pixel[i].Bright, Pixel[i].Fade)
      else begin
        Pixel[i].Bright := SatAdd(Pixel[i].Bright, 256 - Pixel[i].Fade);
        if Pixel[i].Bright = 255 then Pixel[i].Fade := 256 - Pixel[i].Fade;
      end;
end;

procedure TScript.DbgEnd;
begin
  RunTimer.Enabled := false;
  TimerWait := 0;
  ToStr := 0;
  //DebugMode(false);
end;

procedure TScript.SetRunSpeed;
begin
  case RunSpeed.Position of
  0 : RunTimer.Interval := 1000;
  1 : RunTimer.Interval := 500;
  2 : RunTimer.Interval := 200;
  3 : RunTimer.Interval := 100;
  else
    RunTimer.Interval := 1;
  end;
end;

function TScript.ColorToText(aColor: TColor): string;
begin
  Result := IntToHex(aColor and $FF, 2) +
            IntToHex((aColor shr 8) and $FF, 2) +
            IntToHex((aColor shr 16) and $FF, 2);
end;

procedure TScript.DebugAutoIndicator;
var
  St : string;
begin
  inc(DbgInd);
  if DbgInd > 3 then DbgInd := 0;
  case DbgInd of
  0: St := '—';
  1: St := '\';
  2: St := '|';
  3: St := '/';
  end;
  StatusRun(St);
end;

procedure TScript.RunTimerTimer(Sender: TObject);
begin
  DebugAutoIndicator;
   if TimerWait >= 0
    then begin
      DbgForm.WaitDbg;
      Fade;
      dec(TimerWait);
      if(TimerWait < 0) then SelectStr(CurStr);
      UpdateDbgViews;
      exit;
    end
   else ExecCmd(DebugStep);

  if CurStr = ToStr then begin
    RunTimer.Enabled := false;
    TimerWait := 0;
    ToStr := 0;
  end;

  UpdateDbgViews;

  if TimerWait <> 0 then begin
    dec(CurStr);
    SelectStr(CurStr);
    inc(CurStr);
  end
  else SelectStr(CurStr);
end;

procedure TScript.StatusShort(const S : string);
begin
  StatusBar.Panels[0].Text := S;
  if not RunTimer.Enabled then StatusRun('');
end;

procedure TScript.StatusLong(const S : string);
begin
  StatusBar.Panels[2].Text := S;
end;

procedure TScript.StatusRun(const S: string);
begin
  StatusBar.Panels[1].Text := S;
end;

procedure TScript.ScriptEditorSpecialLineColors(Sender: TObject; Line: integer;
  var Special: boolean; var FG, BG: TColor);
begin
  if Line = CurStr then begin
    Special := true;
    BG := clBlue;
    FG := clYellow;
  end else if Line = ToStr then begin
    Special := true;
    BG := clRed;
    FG := clYellow;
  end;
end;

procedure TScript.ActDbgStartStopExecute(Sender: TObject);
begin
  DebugMode(not DbgMode);
end;

procedure TScript.ActAutoRunExecute(Sender: TObject);
begin
  SetRunSpeed;
  RunTimer.Enabled := DbgMode;
end;

procedure TScript.ActColorExecute(Sender: TObject);
begin
  //ColorPick.Visible := not ColorPick.Visible;
  //Splitter.Visible := ColorPick.Visible;
  if ColorDlg.Execute
    then ScriptEditor.InsertTextAtCaret(ColorToText(ColorDlg.SelectedColor));
end;

procedure TScript.ActDbgStepExecute(Sender: TObject);
begin
  ExecCmd(DebugStep);
  UpdateDbgViews;
  SelectStr(CurStr);
end;

procedure TScript.ActDbgStopExecute(Sender: TObject);
begin
  RunTimer.Enabled := false;
  TimerWait := 0;
  ToStr := 0;
  ScriptEditor.Invalidate;
end;

procedure TScript.SelectStr(SN : integer);
begin
  ScriptEditor.CaretY := SN;
  ScriptEditor.MoveCaretToVisibleArea;
  ScriptEditor.Repaint;
end;

procedure TScript.UpdateDbgViews;
begin
  if DbgMode then begin
    DbgForm.UpdateDbg;
  end;
  DbgForm.Visible := DbgMode;
end;

function TScript.SatAdd(X, Y: integer): integer;
begin
  Result := X + Y;
  if Result > 255 then Result := 255;
end;

function TScript.SatSub(X, Y: integer): integer;
begin
  Result := X - Y;
  if Result < 0 then Result := 255;
end;

function TScript.SatMul(X, Y: integer): integer;
begin
  Result := X * Y;
  if Result > 255 then Result := 255;
end;

procedure TScript.PushStack(aStr: integer; aPos : integer = 0);
begin
  if StackHead < cStackDepth-1 then begin
    Stack[StackHead].NumStr := aStr;
    Stack[StackHead].PosStr := aPos;
    inc(StackHead);
  end;
end;

procedure TScript.PushStack(aST: TStackRec);
begin
  PushStack(aST.NumStr, aST.PosStr);
end;

const
  ZeroRec : TStackRec = (NumStr:0;PosStr:0);

function TScript.PopStack: TStackRec;
begin
   if StackHead > 0 then begin
     dec(StackHead);
     Result := Stack[StackHead];
   end
   else
     Result := ZeroRec;
end;

procedure TScript.DebugMode(const Enable : boolean);
begin
   DbgMode := Enable;
   ScriptEditor.ReadOnly := DbgMode;
   //ScriptEditor.Enabled := not DbgMode;
   if(Enable) then begin
     CurStr := 1;
     SelectStr(CurStr);
   end else begin
     DbgEnd;
     CurStr := 0;
     ScriptEditor.HideSelection := true;
     ScriptEditor.Repaint;
   end;
   UpdateDbgViews;
end;

function TScript.DebugStep: string;
begin
   try
     Result := ScriptEditor.Lines[CurStr-1];
     inc(CurStr);
//     SelectStr(CurStr);
   except
     Result := '';
     DebugMode(false);
   end;
   if(CurStr > ScriptEditor.Lines.Count) then begin
     DebugMode(false);
     Result := '';
   end;
end;

function TScript.GetByte(const Cmd: string): integer;
var
  S : string;
  n1, n2 : integer;
begin
  Result := -1;
  try
    S := Copy(Cmd, 1, 2);
    if S = '' then exit;
    if S = 'TP' then Result := cTotalPix-1
    else if S = 'RD' then Result := Random(256)
    else if S = 'RP' then Result := Random(32)
    else if S[1] = 'V' then begin
       n1 := GetNum(S[2]);
       if n1 < 0 then Result := -1
       else Result := CurV[n1];
    end else begin
      n1 := GetNum(Cmd[1]);
      n2 := GetNum(Cmd[2]);
      if (n1 < 0) or (n2 < 0)
        then Result := -1
        else Result := n1 * 16 + n2;
    end;
  except
    Result := -1;
  end;
end;

function TScript.GetNum(const Ch: Char): integer;
begin
  try
    Result := StrToInt('$' + Ch);
  except
    Result := -1;
  end;
end;

function TScript.GetColor(const Cmd: string): Integer;
var
  R, G, B : Integer;
begin
  R := GetByte(Copy(Cmd, 1, 2));
  G := GetByte(Copy(Cmd, 3, 2));
  B := GetByte(Copy(Cmd, 5, 2));
  if (R < 0) or (G < 0) or (B < 0)
    then Result := -1
    else Result := R + G*256 + B*256*256;
end;

procedure TScript.ExecCmd(S : string);
var
  TC : TColor;
  i, D, D1 : integer;
  Str, S1 : string;
  ST : TStackRec;
  po : integer;
label m1;
begin
  Str := S;
  po := 0;
  if CurPos <> 0 then begin
     Str := copy(S, CurPos, 255);
     CurPos := 0;
  end;

  repeat
    Str := trim(Str);
    if Str = '' then break;
    repeat
      try
        if Pos('END', Str) = 1 then begin
          DebugMode(false);
          exit;
        end else if Pos('PC=', Str) = 1 then begin
          // Paint color
          TC := GetColor(Copy(Str, 4, 6));
          if TC < 0
            then break
            else CurColor := TC;
        end else if Pos('GC=', Str) = 1 then begin
          // Global color
          TC := GetColor(Copy(Str, 4, 6));
          if TC < 0
            then break
            else for i := 0 to cTotalPix-1 do Pixel[i].Color := TC;
        end else if Pos('GF=', Str) = 1 then begin
          // Global fade
          D := GetByte(Copy(Str, 4, 2));
          if D < 0
            then break
            else for i := 0 to cTotalPix-1 do Pixel[i].Fade := D;

        end else if Pos('PF=', Str) = 1 then begin
          // Paint fade
          D := GetByte(Copy(Str, 4, 2));
          if D < 0
            then break
            else CurFade := D;
        end else if Pos('PB=', Str) = 1 then begin
          // Paint Bright
          D := GetByte(Copy(Str, 4, 2));
          if D < 0
            then break
            else CurBright := D;
        end else if Pos('GB=', Str) = 1 then begin
          // Global bright
          D := GetByte(Copy(Str, 4, 2));
          if D < 0
            then break
            else for i := 0 to cTotalPix-1 do Pixel[i].Bright := D;
        end else if (Str[1] = 'V') and (Str[3] = '=') then begin
          // Set variable
          D := GetNum(Str[2]);
          D1 := GetByte(Copy(Str, 4, 2));
          if (D < 0) or (D > (cTotalVar-1)) or (D1 < 0)
            then break
            else CurV[D] := D1;

        end else if (Str[1] = 'V') and (Str[3] = '+') then begin
          // Add variable
          D := GetNum(Str[2]);
          D1 := GetByte(Copy(Str, 4, 2));
          if (D < 0) or (D > (cTotalVar-1)) or (D1 < 0)
            then break
            else CurV[D] := SatAdd(CurV[D],D1);

        end else if (Str[1] = 'V') and (Str[3] = '-') then begin
          // Sub variable
          D := GetNum(Str[2]);
          D1 := GetByte(Copy(Str, 4, 2));
          if (D < 0) or (D > (cTotalVar-1)) or (D1 < 0)
            then break
            else CurV[D] := SatSub(CurV[D],D1);

        end else if (Str[1] = 'V') and (Str[3] = '*') then begin
          // Mul variable
          D := GetNum(Str[2]);
          D1 := GetByte(Copy(Str, 4, 2));
          if (D < 0) or (D > (cTotalVar-1)) or (D1 < 0)
            then break
            else CurV[D] := SatMul(CurV[D],D1);

        end else if (Str[1] = 'V') and (Str[3] = '/') then begin
          // Div variable
          D := GetNum(Str[2]);
          D1 := GetByte(Copy(Str, 4, 2));
          if (D < 0) or (D > (cTotalVar-1)) or (D1 < 0)
            then break
            else CurV[D] := CurV[D] div D1;

        end else if (Str[1] = 'V') and (Str[3] = '%') then begin
          // Mod variable
          D := GetNum(Str[2]);
          D1 := GetByte(Copy(Str, 4, 2));
          if (D < 0) or (D > (cTotalVar-1)) or (D1 < 0)
            then break
            else CurV[D] := CurV[D] mod D1;

        end else if Pos('CLR', Str) = 1 then begin
          // Clear
          CurColor := 0;
          CurFade := 0;
          CurBright := 255;
          StackHead := 0;
          for D := 0 to (cTotalVar-1) do CurV[D] := 0;
          for D := 0 to cTotalPix-1 do begin
              Pixel[D].Bright := 0;
              Pixel[D].Fade := 0;
              Pixel[D].Color := 0;
          end;

        end else if Pos('NEG', Str) = 1 then begin
          // Inverse bright
          for D := 0 to cTotalPix-1 do Pixel[D].Bright := Pixel[D].Bright xor 255;

        end else if Pos('REV', Str) = 1 then begin
          // Reverse
          Rev := not Rev;

        end else if Pos('RPT', Str) = 1 then begin
          // Repeat begin
          // CurPos := pos(S, Str);
          if length(Str) > 3
            then PushStack(CurStr - 1, po+3)
            else PushStack(CurStr);

        end else if Pos('RST', Str) = 1 then begin
          // Restart
          CurStr := 1;
          //ScriptEditor.Repaint;
          SelectStr(CurStr);

        end else if Pos('LV', Str) = 1 then begin
          // Loop until
          D1 := GetByte(copy(Str, 5, 2));
          D := GetNum(Str[3]);
          if (D < 0) or (D > (cTotalVar-1)) then exit;
          // проверка на цикл по условию
          S1 := copy(Str, 4, 1);
          if (S1 <> '') and (S1[1] in ['=','>','<','!']) then begin
             case copy(Str, 4, 1) of
             '=' : if CurV[D] = D1 then goto m1;
             '>' : if CurV[D] > D1 then goto m1;
             '<' : if CurV[D] < D1 then goto m1;
             '!' : if CurV[D] <> D1 then goto m1;
             end;
             PopStack;
             break;
          end;

          if CurV[D] <> 0 then begin
             CurV[D] := SatSub(CurV[D], 1);
        m1:
            ST := PopStack;
            if ST.NumStr = 0 then ST.NumStr :=1; // обработка конца цикла без начала
            if CurStr = ST.NumStr then begin
              Str := S;
              PushStack(ST);
            end else begin
              CurStr := ST.NumStr;
              CurPos := ST.PosStr;
              PushStack(CurStr, CurPos);
              SelectStr(CurStr);
              Str := '';
            end;
          end
          else PopStack;

        end else if Pos('INF', Str) = 1 then begin
          // Infinity loop
          goto m1;

        end else if Pos('WT=', Str) = 1 then begin
          // Wait
          D := GetByte(Copy(Str, 4, 2));
          if D < 0
            then break
            else Wait(D);

        end else if Pos('PNT', Str) = 1 then begin
          // Paint
          Wait(1);

        end else if Pos('PM=', Str) = 1 then begin
          // Paint map
          i := 4;
          D := GetByte(Copy(Str, i, 2));
          while (D in [0..255]) do begin
            if D <= cTotalPix-1 then begin
               if Rev then D := cTotalPix-1 - D;
               Pixel[D].Fade := CurFade;
               Pixel[D].Bright := CurBright;
               Pixel[D].Color := CurColor;
            end;
            inc(i, 2);
            D := GetByte(Copy(Str, i, 2));
          end;
        end;
      except
      end;
    until true;
    while (Str <> '') and (not (Str[1] in [' ', #$0d, #$0a, #$09]))
      do begin
        delete(Str, 1, 1);
        inc(po);
      end;
    while (Str <> '') and (Str[1] in [' ', #$0d, #$0a, #$09])
      do begin
        delete(Str, 1, 1);
        inc(po);
      end;
  until Str = '';
end;

end.

